import json
import math
import time

import pandas as pd

from typing import List, Union
from pycollimator.api import Api
from pycollimator.diagrams import Block
from pycollimator.i18n import N
from pycollimator.log import Log
from pycollimator.models import Model, ModelParameters, load_model
from pycollimator.simulation_hashed_file import SimulationHashedFile
from pycollimator.utils import is_uuid
from pycollimator.results import SimulationResults, LinearizationResult


class SimulationLogLine:
    def __init__(self, data):
        self._data = data
        self._str: str = ""
        self._parse(data)

    def __str__(self) -> str:
        return self._str

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._data}>"

    def _parse(self, data):
        try:
            js_data = data
            if isinstance(data, str):
                js_data = json.loads(data)
            ts = ""
            if js_data.get("timestamp") is not None:
                ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(js_data["timestamp"])) + " "
            parsed_line = f"{ts}{js_data.get('level')} {js_data.get('message')}"
            for k, v in js_data.items():
                if k not in ["timestamp", "level", "message"]:
                    parsed_line += f" {k}={v}"
            self._str = parsed_line
        except Exception as e:
            Log.trace(f"Failed to parse log line: {e}")
            self._str = str(data)


class Simulation:
    def __init__(self, data: dict, model: Model):
        self.model = model
        self._data = data
        self._logs: List[SimulationLogLine] = None
        self._results: pd.DataFrame = None

    def __getitem__(self, key):
        return self._data[key]

    def __repr__(self) -> str:
        if Log.is_level_above("DEBUG"):
            return f"<{self.__class__.__name__} status='{self.status}' uuid='{self.uuid}'>"
        return f"<{self.__class__.__name__} status='{self.status}'>"

    @property
    def uuid(self) -> str:
        return self._data["uuid"]

    @property
    # this is last status, not latest as it doesn't request to backend.
    def status(self) -> str:
        return self._data["status"]

    @property
    def model_uuid(self) -> str:
        return self._data["model_uuid"]

    @property
    def results_available(self) -> bool:
        return self._data.get("results_available") is True

    @property
    def is_failed(self) -> bool:
        return self.status == "failed"

    @property
    def is_completed(self) -> bool:
        return self.status == "completed"

    @property
    def is_running(self) -> bool:
        return self.status != "created" and not self.is_failed and not self.is_completed

    @property
    def is_started(self) -> bool:
        return self.status != "created"

    @property
    def logs(self) -> List[SimulationLogLine]:
        if self._logs is None:
            logs = Api.simulation_logs(self.model_uuid, self.uuid)
            self._logs = [SimulationLogLine(s) for s in str(logs).splitlines()]
        return self._logs

    @property
    def compilation_logs(self) -> SimulationLogLine:
        logs = self._data.get("compilation_logs")
        if logs is None:
            return None
        if isinstance(logs, list):
            return SimulationLogLine(logs[0])
        return SimulationLogLine(logs)

    def show_logs(self) -> None:
        # It would be awesome to use markdown to pretty-print these logs
        if self.is_failed:
            print(self.compilation_logs)
            return
        for log_line in self.logs:
            print(log_line)

    @property
    def results(self) -> SimulationResults:
        # TODO: cache continuous results
        if not self.results_available:
            Log.warning("Simulation results may not be available yet:", self)
            # But we will try downloading them anyway since that's what the user asked for
        results = Api.simulation_results(self.model_uuid, self.uuid)
        return SimulationResults(results, self.model)

    def _signal_results(self, paths: List[str]) -> SimulationResults:
        if not self.results_available:
            Log.warning("Simulation results may not be available yet:", self)
            # But we will try downloading them anyway since that's what the user asked for
        results_pds = [
            SimulationResults(result, self.model).to_pandas()
            for result in Api.signal_results(self.model_uuid, self.uuid, paths)
        ]
        merged_results_csv = pd.concat(results_pds, axis=1).to_csv()
        return SimulationResults(merged_results_csv, self.model)

    def get_results(self, signals: Union[List[str], str] = None, wait=True) -> SimulationResults:
        """Get simulation results.

        :param signals: List of signal paths to fetch results of or a single signal path, defaults to None (all signals)
        :type signals: List[str] | str, optional
        :param wait: Whether or not to wait for the simulation to complete, defaults to True
        :type wait: bool, optional
        :return: Simulation results of specified signal, or all results if none specified.
        :rtype: SimulationResults
        """
        if wait:
            self.wait()
        if signals is not None:
            return self._signal_results([signals] if isinstance(signals, str) else signals)
        else:
            return self.results

    def update(self):
        Log.trace("Updating simulation:", self)
        self._data = Api.simulation_get(self.model_uuid, self.uuid)
        return self

    def start(self):
        Log.debug("Starting simulation:", self)
        self._data = Api.simulation_start(self.model_uuid, self.uuid)

    def stop(self):
        Log.debug("Stopping simulation:", self)
        self._data = Api.simulation_stop(self.model_uuid, self.uuid)

    def wait(self):
        if self.is_completed:
            return

        if not self.is_started:
            Log.debug("Simulation not started:", self)
            self.start()

        start_ts = time.time()
        while self.is_running:
            Log.debug(f"Waiting for simulation {math.trunc(time.time() - start_ts)}s:", self)
            self.update()
            if self.status == "completed":
                break
            time.sleep(1)
        Log.debug("Simulation completed:", self)

    def to_pandas(self, wait=True):
        return self.get_results(wait=wait).to_pandas()

    def _upload_hashed_files(self, model, simulation):
        parameter_overrides = []
        blocks = model._get_blocks_with_data()
        for block_uuid in blocks:
            input_data = blocks[block_uuid]
            sdf = SimulationHashedFile(input_data)
            resp = sdf.upload(model.uuid, simulation.uuid)
            Log.trace(f"Uploaded hashed file for block '{block_uuid}':", resp)
            parameter_overrides.append(
                {
                    "expression": f'__hashed_file("{sdf.hash}", "{sdf.content_type}")',
                    "block_uuid": block_uuid,
                    "parameter_name": "file_name",
                }
            )
        Log.trace("Parameter overrides:", parameter_overrides)
        return parameter_overrides


def _run_simulation_common(model, linearization, parameters, wait, no_sync, ignore_cache, snapshot_uuid) -> Simulation:
    Log.trace("run_simulation:", model.__repr__())
    if not isinstance(model, Model):
        model = load_model(model)

    if not no_sync:
        model.sync()

    if linearization is not None:
        if is_uuid(linearization):
            linearization = {"submodel_uuid": linearization}
        elif isinstance(linearization, Block):
            linearization = {"submodel_uuid": linearization.uuid}
        else:
            raise ValueError(N("Invalid argument 'linearization'"))

    # Prepare DataSource files
    hashed_files = model._get_hashed_files()
    # make special case for file name
    overrides = [
        {
            "expression": f'__hashed_file("{hashed_file.hash}", "{hashed_file.content_type}")',
            "block_uuid": block_uuid,
            "parameter_name": "file_name",
        }
        for block_uuid, hashed_file in hashed_files.items()
    ]

    # Send simulation specific parameters
    parameters = {
        **model.parameters.to_api_data(),
        **(ModelParameters(parameters or {}).to_api_data()),
    }

    # Simulation-specific configuration
    configuration = model.configuration.to_dict()

    sim_request = {
        "version": model.version,
        "no_start": True,
        "configuration": configuration,
        "overrides": overrides,
        "parameters": parameters,
        "ignore_cache": ignore_cache,
    }

    if linearization is not None:
        sim_request["linearization"] = linearization

    if snapshot_uuid is not None:
        sim_request["snapshot_uuid"] = snapshot_uuid

    sim = Simulation(Api.simulation_create(model.uuid, sim_request), model)

    if sim.status == "created":
        # Upload DataSource files and then update the BlockParameterOverrides
        parameter_overrides = sim._upload_hashed_files(model, sim)
        if len(parameter_overrides) > 0:
            body = {"overrides": parameter_overrides}
            sim._data = Api.simulation_parameters_set(model.uuid, sim.uuid, body)
        sim.start()

    if wait:
        sim.wait()

    if sim.is_failed:
        Log.error(N("Simulation failed! Use .show_logs() to see the logs."))

    return sim


def run_simulation(
    model, parameters: dict = None, wait=True, no_sync=False, ignore_cache=False, snapshot_uuid=None
) -> Simulation:
    """
    Run a simulation on the given model.
    """
    return _run_simulation_common(model, None, parameters, wait, no_sync, ignore_cache, snapshot_uuid)


def linearize(model, submodel, parameters: dict = None, no_sync=False, ignore_cache=False, snapshot_uuid=None):
    """
    Linearize the submodel within a model.
    """
    if not isinstance(model, Model):
        model = load_model(model)

    if not isinstance(submodel, Block):
        submodel = model.find_block(submodel)

    sim = _run_simulation_common(model, submodel, parameters, True, no_sync, ignore_cache, snapshot_uuid)

    if not sim.is_completed:
        Log.error(N("Simulation could not complete!"))
        sim.show_logs()
        return sim

    lin_results_csv = Api.linearization_results_csv(model.uuid, sim.uuid)
    res = LinearizationResult._from_csv(lin_results_csv)
    return res



import re

import pandas as pd
from typing import List, Optional

from pycollimator.error import NotFoundError
from pycollimator.i18n import N
from pycollimator.log import Log


# Current structure of SimulationModel:
# "diagram" : { ModelDiagram }
# "submodels": {
#   "references": {
#      "<uuid>": { "diagram_uuid": "<uuid>" },
#       ...
#   },
#   "diagrams": {
#     "<uuid>": { ModelDiagram }
#   }
# }


# FIXME FIXME FIXME
# We don't have a strong representation of the model, and we create
# Block objects on the fly. Modifying them means we need to backtrack
# to the model and update it somehow. Right now this scenario is limited
# to DataSourceBlock input data. We don't really want to allow modifying
# the model, so it's kinda okay. At least for now.


class Block:
    """
    Representation of a block in a model.

    Can be returned to the API user.
    """

    @classmethod
    def from_data(cls, data, model, path):
        if data.get("type") == "core.DataSource":
            return DataSourceBlock(data, model, path)
        return cls(data, model, path)

    def __init__(self, block_json, model, path):
        self.model = model
        self._json = block_json
        self.path = path

    def __getitem__(self, key):
        return self._json[key]

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        if Log.is_level_above("DEBUG"):
            return f"<{self.__class__.__name__} name='{self.name}' type='{self.type}' uuid='{self.uuid}'>"
        return f"<{self.__class__.__name__} name='{self.name}' type='{self.type}'>"

    @property
    def name(self):
        return self._json["name"]

    @property
    def uuid(self):
        return self._json["uuid"]

    @property
    def type(self):
        return self._json["type"]

    # @property
    # def parameters(self):
    #     return self._json["parameters"]

    # TODO: expose setting block params

    def get_parameter(self, name: str, no_eval=False):
        param = self._json["parameters"].get(name)
        Log.trace(f"get_parameter: {name}={param}")
        if param is None:
            raise NotFoundError(N(f"Block '{self}' of type '{self.type}' does not have parameter '{name}'"))
        if param.get("is_string", False) is True:
            return str(param["value"])
        if no_eval is True:
            return param["value"]
        expr = param.get("expression") or param["value"]
        evaluated = eval(expr)
        return evaluated


# Special block that reads csv
class DataSourceBlock(Block):
    def __init__(self, block_json, model, path):
        if block_json.get("type") != "core.DataSource":
            raise TypeError(N("DataSourceBlock must be created from a DataSource block"))
        super().__init__(block_json, model, path)
        self.data = None

    def set_data(self, data: pd.DataFrame):
        # FIXME make sure the shape is correct and all that
        if not isinstance(data, pd.DataFrame):
            raise TypeError(N("Input data must be a pandas DataFrame"))
        # set data of a DataSource block
        Log.trace("set_data, shape:", data.shape, "block:", self.__repr__())
        self.data = data.copy()
        self.data.index.name = "time"
        self.model._set_datasource_data(self, self.data)


class ModelDiagram:
    """
    Contents of a fully loaded model diagram (single plane).

    For use by internal APIs.
    """

    def __init__(self, data, model):
        self.model = model
        self.diagram = data

    def __str__(self) -> str:
        if self.diagram.get("name") is not None:
            return self.diagram["name"]
        return self.diagram["uuid"]

    def __repr__(self) -> str:
        if Log.is_level_above("DEBUG"):
            uid = self.diagram["uuid"]
            return f"<{self.__class__.__name__} model='{self.model}' uuid='{uid}'>"
        return f"<{self.__class__.__name__} model='{self.model}'>"

    @property
    def nodes(self):
        return self.diagram.get("nodes", [])

    @property
    def links(self):
        return self.diagram.get("links", [])

    # Full path only works at the model graph level.
    def find_block(self, pattern: str = None, name: str = None, type: str = None, ignorecase=True) -> Optional[Block]:
        blocks = self.find_blocks(pattern=pattern, name=name, type=type, case=ignorecase)
        if len(blocks) == 0:
            return None
        if len(blocks) > 1:
            raise NotFoundError(N(f"Multiple blocks found for '{name}' in model '{self}'"))
        return blocks[0]

    # todo: do a search on model graph that constructs path 5
    # Not to be called with a path name as path name is unique.
    # probably won't need this ModelDiagram helper once BFS / path construction implemented at ModelGraph level
    def find_blocks(self, pattern: str = None, name: str = None, type: str = None, case=True) -> List[Block]:
        found = None
        model_graph = self.model._graph

        if name is None and pattern is None and type is None:
            pattern = ""  # matches any string

        if pattern is not None:
            rgx = re.compile(pattern, re.IGNORECASE if not case else 0)
            # yes. yikes. but going away with support for V2
            found = [
                Block.from_data(
                    node, self.model, model_graph._get_block_path(model_graph.root_diagram, node.get("uuid"), "")
                )
                for node in self.nodes
                if rgx.match(node.get("name"))
            ]
        if name is not None:
            if not case:
                found = [
                    Block.from_data(
                        node, self.model, model_graph._get_block_path(model_graph.root_diagram, node.get("uuid"), "")
                    )
                    for node in self.nodes
                    if node.get("name", "").lower() == name.lower()
                ]
            else:
                found = [
                    Block.from_data(
                        node, self.model, model_graph._get_block_path(model_graph.root_diagram, node.get("uuid"), "")
                    )
                    for node in self.nodes
                    if node.get("name") == name
                ]

        # If type is set: filter by type, or return all blocks of the given type if no other search criteria was set
        if type is not None:
            if not type.startswith("core."):
                type = "core." + type
            type = type.lower()
            if name is None and pattern is None:
                found = [
                    Block.from_data(
                        node, self.model, model_graph._get_block_path(model_graph.root_diagram, node.get("uuid"), "")
                    )
                    for node in self.nodes
                    if node.get("type").lower() == type
                ]
            else:
                found = [blk for blk in found if blk.type.lower() == type]

        return found


class ModelGraph:
    """
    Contents of a fully loaded model graph (all loadable planes).

    For use by internal APIs.
    """

    def __init__(self, data, model):
        self._data = data
        self._model = model
        self.uuid = data.get("uuid")
        self.name = data.get("name")
        self.root_diagram = ModelDiagram(data["diagram"], model=self._model)
        self.diagrams_by_submodel_uuid = {}  # type: Dict[str, ModelDiagram]

        submodel_diagrams = data.get("submodels", {}).get("diagrams", {})
        submodel_references = data.get("submodels", {}).get("references", {})
        self.submodel_references = submodel_references

        for submodel_uuid in self.submodel_references:
            ref = self.submodel_references[submodel_uuid]
            diagram_uuid = ref["diagram_uuid"]
            diagram = submodel_diagrams[diagram_uuid]
            self.diagrams_by_submodel_uuid[submodel_uuid] = ModelDiagram(diagram, model)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        if Log.is_level_above("DEBUG"):
            return f"<{self.__class__.__name__} model='{self._model}' uuid='{self.uuid}'>"
        return f"<{self.__class__.__name__} model='{self._model}'>"

    # def get_diagram(self, diagram_uuid: str = None):
    #     if diagram_uuid is None:
    #         return self.root_diagram
    #     return ModelDiagram(self._data["diagrams"][diagram_uuid], model=self._model)

    # def get_submodel_diagram(self, submodel_uuid):
    #     return ModelDiagram(
    #         self.get_diagram(self._data["submodels"]["references"][submodel_uuid]["diagram_uuid"]), model=self._model
    #     )

    # todo add case flag.
    # todo generalize
    def find_block_by_path(self, path: str):
        diagram = self.root_diagram
        path_to_traverse = path.split(".")

        for name in path_to_traverse[:-1]:
            matched_nodes = [node for node in diagram.nodes if node["name"] == name]
            if len(matched_nodes) == 0:
                raise NotFoundError(N(f"No block was found in path '{path}' "))
            if len(matched_nodes) > 1:
                raise NotFoundError(N(f"Multiple matching blocks found in path '{path}' "))
            node = matched_nodes[0]
            if not (node["type"] == "core.Submodel" or node["type"] == "core.Group"):
                raise NotFoundError(N(f"No block was found in path '{path}' "))
            # move onto next diagram
            diagram = self.diagrams_by_submodel_uuid[node["uuid"]]

        # last name in path can be any type of block
        last_name = path_to_traverse[-1]
        found_blocks = [node for node in diagram.nodes if node["name"] == last_name]

        if len(found_blocks) == 0:
            raise NotFoundError(N(f"No block was found in path '{path}' "))
        if len(found_blocks) > 1:
            raise NotFoundError(N(f"Multiple matching blocks found in path '{path}' "))

        return Block.from_data(found_blocks[0], diagram.model, path)

    # FIXME walk and construct blocks paths
    # For non-path search.
    # The ModelGraph is the 'root' class in which you can traverse. ModelDiagram is a node.
    # No cycles in graph/diagrams representation of model.
    # With submodels v2, UUIDs are not unique, so we must construct full paths for any block that is initialized.
    # Later, consider doing this upfront and only constructing full graph once.
    # todo: do a search the model graph that constructs path 4.
    def find_blocks(self, pattern: str = None, name: str = None, type: str = None, case=True) -> List[Block]:
        found = []

        # find in root diagram
        nodes = self.root_diagram.find_blocks(pattern=pattern, name=name, type=type, case=case)
        for node in nodes:
            found.append(node)

        # todo: BFS through submodel and construct paths
        # FIXME: this walks submodels in random order and doesn't know blocks paths
        # need to use the references
        for submodel_uuid in self.diagrams_by_submodel_uuid:
            submodel_diagram = self.diagrams_by_submodel_uuid[submodel_uuid]
            nodes = submodel_diagram.find_blocks(pattern=pattern, name=name, type=type, case=case)
            for node in nodes:
                found.append(node)

        return found

    # ! Supports V1 only where UUID is unique. OK for now as there's no V2 support yet
    # (submodels missing in V2 models), but will go away with path construction traversal
    # todo remove function completely for V2 submodels as UUID is not unique
    def _get_block_path(self, diagram: ModelDiagram, block_uuid: str, pfx: str) -> str:
        for node in diagram.nodes:
            if node["uuid"] == block_uuid:
                return f"{pfx}{node['name']}"
            if node["type"] == "core.Submodel" or node["type"] == "core.Group":
                submodel_diagram = self.diagrams_by_submodel_uuid[node["uuid"]]
                found = self._get_block_path(submodel_diagram, block_uuid, f"{pfx}{node['name']}.")
                if found:
                    return found
        return None

    def get_block_path(self, block):
        """
        Get the path of a block in the model.
        """
        if not isinstance(block, Block):
            blocks = self.find_blocks(name=block)
            if len(blocks) == 0:
                blocks = self.find_blocks(pattern=block)
            if len(blocks) == 0:
                raise NotFoundError(N(f"No block found for '{block}'"))
            if len(blocks) > 1:
                raise NotFoundError(N(f"Multiple blocks found for '{block}'"))
            return blocks[0].path
        return block.path



# Simple script to run a model and get the results
import pycollimator as collimator

# Your dashboard generated token
token = ""
# Project uuid
project = ""
# Api url of the environment your project is in
api_url = "https://dev.collimator.ai"
# Unique model name from the project
model_name = ""


# Set the global variables used for auth
collimator.set_auth_token(token, project, api_url)

# Fetch model
model = collimator.load_model(model_name)

# Set model parameters for the simulation. Not saved to model by default.
model.set_parameters({"parameter1": 1, "parameter2": 2})

# Set model configuration for the simulation. Not saved to model by default.
model.set_configuration({"stop_time": 10})

# TODO: add find_block example

# Run the simulation
sim = collimator.run_simulation(model, wait=True)

# Get the results
results = sim.results

# Access the results as a pandas dataframe using block path
results.to_pandas()["Submodel_0.Plant_1.out_0"]

# TODO: add results[] example



#!/bin/env python3

"""Setup script for collimator notebook package."""

from setuptools import setup, find_packages  # type: ignore

deps = ["numpy", "pandas", "requests", "aiohttp", "control", "simpleeval", "ipywidgets==7.7.2"]

setup(
    name="pycollimator",
    setup_requires=["setuptools_scm"],
    use_scm_version={
        "root": "../../../",
        "relative_to": __file__,
        "local_scheme": "no-local-version",
        "version_scheme": "post-release",
    },
    packages=["pycollimator"],
    python_requires=">=3.7",
    install_requires=deps,
    long_description="Python package for Collimator.ai",
)
